@@ -118,11 +118,18 @@ typedef enum
 #endif
 #define FATAL		21			/* fatal error - abort process */
 #define PANIC		22			/* take down the other backends with me */
+/* pgpool-II extension. This is same as ERROR but sets the
+ * do not cache connection flag before transforming to ERROR.
+ */
+#define FRONTEND_ERROR	23			/* transformed to ERROR at errstart */
 
  /* #define DEBUG DEBUG1 */	/* Backward compatibility with pre-7.3 */
-#define POOL_EXIT_SUCCESS	0	/* failure exit and child gets restarted*/
-#define POOL_EXIT_NOFATAL	1	/* failure exit and child gets restarted*/
-#define POOL_EXIT_FATAL		3	/* This exit code from child takes down the pgpool main with it */
+#define POOL_EXIT_NO_RESTART	0	/* child exiting with this error will not be
+									 * restarted by pgpool main process
+									 * but the pgpool main will remain unaffected by this */
+#define POOL_EXIT_AND_RESTART	1	/* child process exit with this error code will
+									 * be restarted by the pgpool main */
+#define POOL_EXIT_FATAL			3	/* This exit code from child also takes down the pgpool main with it */
 
 
 
@@ -280,6 +287,7 @@ extern int	errposition(int cursorpos);
 
 #define pg_unreachable() exit(0)
 
+extern bool  getfrontendinvalid(void);
 extern int	geterrcode(void);
 extern int	geterrposition(void);
 extern int	getinternalerrposition(void);
@@ -431,6 +439,7 @@ typedef struct ErrorData
 	const char *domain;			/* message domain */
 	const char *context_domain; /* message domain for context message */
 	int			sqlerrcode;		/* encoded ERRSTATE */
+	bool		frontend_invalid;/* true when frontend connection is not valid */
 	char	   *pgpool_errcode;	/* error code to be sent to client */
 	char	   *message;		/* primary error message */
 	char	   *detail;			/* detail error message */
