@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2014	PgPool Global Development Group
+ * Copyright (c) 2003-2016	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -45,7 +45,6 @@
 
 #include "pool.h"
 #include "rewrite/pool_timestamp.h"
-#include "parallel_query/pool_rewrite_query.h"
 #include "rewrite/pool_lobj.h"
 #include "protocol/pool_proto_modules.h"
 #include "pool_config.h"
@@ -76,7 +75,6 @@ static int allow_close_transaction = 1;
 
 int is_select_pgcatalog = 0;
 int is_select_for_update = 0; /* 1 if SELECT INTO or SELECT FOR UPDATE */
-bool is_parallel_table = false;
 
 /*
  * last query string sent to simpleQuery()
@@ -89,7 +87,6 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 									 POOL_CONNECTION_POOL *backend,
 									 POOL_SENT_MESSAGE *message);
 static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *number_of_nodes);
-static int extract_ntuples(char *message);
 static POOL_STATUS close_standby_transactions(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend);
 /*
@@ -116,6 +113,8 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
+	bool error;
+
 	/* Get session context */
 	session_context = pool_get_session_context(false);
 
@@ -173,12 +172,12 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	MemoryContext old_context = MemoryContextSwitchTo(query_context->memory_context);
 
 	/* parse SQL string */
-	parse_tree_list = raw_parser(contents);
+	parse_tree_list = raw_parser(contents, &error);
 
 	if (parse_tree_list == NIL)
 	{
 		/* is the query empty? */
-		if (*contents == '\0' || *contents == ';')
+		if (*contents == '\0' || *contents == ';' || error == false)
 		{
 			/*
 			 * JBoss sends empty queries for checking connections.
@@ -186,7 +185,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			 * to affect load balance.
 			 * [Pgpool-general] Confused about JDBC and load balancing
 			 */
-			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY, &error);
 		}
 		else
 		{
@@ -209,7 +208,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 				ereport(LOG,
 						(errmsg("Unable to parse the query: \"%s\" from client %s(%s)", contents, remote_host, remote_port)));
 			}
-			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY, &error);
 			query_context->is_parse_error = true;
 		}
 	}
@@ -254,19 +253,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			query_context->is_multi_statement = false;
 		}
 
-		if (PARALLEL_MODE)
-		{
-			bool parallel = true;
-
-			is_parallel_table = is_partition_table(backend,node);
-			status = pool_do_parallel_query(frontend, backend, node, &parallel, &contents, &len);
-			if (parallel)
-			{
-				pool_query_context_destroy(query_context);
-				return status;
-			}
-		}
-
 		/* check COPY FROM STDIN
 		 * if true, set copy_* variable
 		 */
@@ -613,7 +599,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
         ereport(FATAL,
                 (return_code(2),
                  errmsg("unable to Execute"),
-                 errdetail("unable to get bind message")));
+                 errdetail("unable to get query context")));
 
 	if (!bind_msg->query_context->original_query)
         ereport(FATAL,
@@ -730,7 +716,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/* check if query is "COMMIT" or "ROLLBACK" */
 	commit = is_commit_or_rollback_query(node);
 
-	if (REPLICATION || PARALLEL_MODE)
+	if (!STREAM)
 	{
 		/*
 		 * Query is not commit/rollback
@@ -738,7 +724,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		if (!commit)
 		{
 			/* Send the query to master node */
-			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, false);
 
 			/* Check specific errors */
 			specific_error = check_errors(backend, MASTER_NODE_ID);
@@ -757,27 +743,55 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			memset(msg + len, 0, sizeof(int));
 
 			/* send query to other nodes */
-			pool_extended_send_and_wait(query_context, "E", len, msg, -1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, msg, -1, MASTER_NODE_ID, false);
 		}
 		else
 		{
-			pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID, false);
 		}
 		
 		/* send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
 		if (commit)
 		{
-			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, false);
 		}
 	}
-	else
+	else		/* streaming replication mode */
 	{
-		pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID);
-		pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID);
+		pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID, true);
+
+#ifdef NOT_USED
+		/*
+		 * Send flush message to backend to make sure that we get any response
+		 * from backend.
+		 */
+		pool_write(MASTER(session_context->backend), "H", 1);
+		len = htonl(sizeof(len));
+		pool_write_and_flush(MASTER(session_context->backend), &len, sizeof(len));
+
+		if (MASTER(session_context->backend)->db_node_id != session_context->load_balance_node_id)
+		{
+			POOL_CONNECTION *con;
+
+			con = session_context->backend->slots[session_context->load_balance_node_id]->con;
+			pool_write(con, "H", 1);
+			len = htonl(sizeof(len));
+			pool_write_and_flush(con, &len, sizeof(len));
+		}
+#endif
+		/*
+		 * Remeber that we send flush or sync message to backend.
+		 */
+		pool_unset_pending_response();
+#ifdef NOT_USED
+		pool_unset_query_in_progress();
+#endif
 	}
 
 	return POOL_CONTINUE;
 }
+
 /*
  * process Parse (V3 only)
  */
@@ -795,6 +809,8 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
+	bool error;
+
 	/* Get session context */
 	session_context = pool_get_session_context(false);
 
@@ -809,12 +825,12 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	/* parse SQL string */
 	MemoryContext old_context = MemoryContextSwitchTo(query_context->memory_context);
-	parse_tree_list = raw_parser(stmt);
+	parse_tree_list = raw_parser(stmt, &error);
 
 	if (parse_tree_list == NIL)
 	{
 		/* is the query empty? */
-		if (*stmt == '\0' || *stmt == ';')
+		if (*stmt == '\0' || *stmt == ';' || error == false)
 		{
 			/*
 			 * JBoss sends empty queries for checking connections.
@@ -822,7 +838,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			 * to affect load balance.
 			 * [Pgpool-general] Confused about JDBC and load balancing
 			 */
-			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY, &error);
 		}
 		else
 		{
@@ -845,7 +861,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				ereport(LOG,
 						(errmsg("Unable to parse the query: \"%s\" from client %s(%s)", stmt, remote_host, remote_port)));
 			}
-			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY, &error);
 			query_context->is_parse_error = true;
 		}
 	}
@@ -1042,7 +1058,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 	}
 
-	if (REPLICATION || PARALLEL_MODE || MASTER_SLAVE)
+	if (REPLICATION || SLONY)
 	{
 		/*
 		 * We must synchronize because Parse message acquires table
@@ -1050,7 +1066,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		 */
 		ereport(DEBUG1,
 				(errmsg("Parse: waiting for master completing the query")));
-        pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID);
+        pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID, false);
 
 
 		/*
@@ -1091,12 +1107,22 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 		else
 		{
-			pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID, false);
 		}
 	}
+	else if (STREAM)
+	{
+		/* XXX fix me:even with streaming replication mode, couldn't we have a deadlock */
+		pool_set_query_in_progress();
+		pool_clear_sync_map();
+		pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID, true);
+		pool_add_sent_message(session_context->uncompleted_message);
+		pool_unset_query_in_progress();
+	}
 	else
 	{
-		pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID);
+		pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID, false);
 	}
 
 	return POOL_CONTINUE;
@@ -1114,6 +1140,7 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 	int insert_stmt_with_lock = 0;
+	bool nowait;
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
@@ -1165,6 +1192,10 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	session_context->query_context = query_context;
 
+/*
+ * Fix me
+ */
+#ifdef NOT_USED
 	if (pool_config->load_balance_mode && pool_is_writing_transaction())
 	{
 		pool_where_to_send(query_context, query_context->original_query,
@@ -1173,6 +1204,7 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		if (parse_before_bind(frontend, backend, parse_msg) != POOL_CONTINUE)
 			return POOL_END;
 	}
+#endif
 
 	/*
 	 * Start a transaction if necessary in replication mode
@@ -1204,12 +1236,23 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			}
 		}
 	}
+
 	ereport(DEBUG1,
 		(errmsg("Bind: waiting for master completing the query")));
 
-	pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID);
+	pool_set_query_in_progress();
 
-	pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID);
+	nowait = (STREAM? true: false);
+
+	pool_clear_sync_map();
+	pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID, nowait);
+	pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID, nowait);
+
+	if (STREAM)
+	{
+		pool_unset_query_in_progress();
+		pool_add_sent_message(session_context->uncompleted_message);
+	}
 	
 	if(rewrite_msg)
 		pfree(rewrite_msg);
@@ -1223,6 +1266,8 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
+	bool nowait;
+
 	/* Get session context */
 	session_context = pool_get_session_context(false);
 
@@ -1268,9 +1313,14 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
     ereport(DEBUG1,
             (errmsg("Describe: waiting for master completing the query")));
 
-	pool_extended_send_and_wait(query_context, "D", len, contents, 1, MASTER_NODE_ID);
+	nowait = (STREAM? true: false);
 
-	pool_extended_send_and_wait(query_context, "D", len, contents, -1, MASTER_NODE_ID);
+	pool_set_query_in_progress();
+	pool_extended_send_and_wait(query_context, "D", len, contents, 1, MASTER_NODE_ID, nowait);
+	pool_extended_send_and_wait(query_context, "D", len, contents, -1, MASTER_NODE_ID, nowait);
+
+	if (STREAM)
+		pool_unset_query_in_progress();
 
 	return POOL_CONTINUE;
 }
@@ -1303,9 +1353,10 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
                 (return_code(2),
                     errmsg("unable to execute close, invalid message")));
 	/* 
-	 * As per the postgresql, calling close on non existing portals is not
-	 * an error. So on the same footings we will ignore all such calls and
-	 * return the close complete message to clients with out going to backend
+	 * As per the postgresql, calling close on non existing portals or
+	 * statements is not an error. So on the same footings we will ignore all
+	 * such calls and return the close complete message to clients with out
+	 * going to backend
 	 */
 	if (!msg)
 	{
@@ -1333,9 +1384,26 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
     ereport(DEBUG1,
             (errmsg("Close: waiting for master completing the query")));
-	pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID);
 
-	pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID);
+	pool_set_query_in_progress();
+
+	if (!STREAM)
+	{
+		pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID, false);
+		pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID, false);
+	}
+	else
+	{
+		POOL_PENDING_MESSAGE *pmsg;
+
+		pool_clear_sync_map();
+		pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID, true);
+
+		pmsg = pool_pending_messages_create('C', len, contents);
+		pool_pending_message_add(pmsg);
+		pool_unset_query_in_progress();
+	}
 
 	return POOL_CONTINUE;
 }
@@ -1381,9 +1449,19 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	POOL_SESSION_CONTEXT *session_context;
 	Node *node = NULL;
 	char *query = NULL;
+	POOL_SYNC_MAP_STATE use_sync_map;
+
+	/*
+	 * It is possible that the "ignore until sync is received" flag was set if
+	 * we send sync to backend and the backend returns error. Let's reset the
+	 * flag unconditionally because we apparently have received a "ready for
+	 * query" message from backend.
+	 */
+	pool_unset_ignore_till_sync();
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
+	use_sync_map = pool_use_sync_map();
 
 	/*
 	 * If the numbers of update tuples are differ and
@@ -1442,7 +1520,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				free_string(msg);
 
 				degenerate_backend_set(victim_nodes, number_of_nodes);
-				child_exit(1);
+				child_exit(POOL_EXIT_AND_RESTART);
 			}
 			else
 			{
@@ -1521,9 +1599,14 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
-			if (!VALID_BACKEND(i))
+			if (!VALID_BACKEND(i) || use_sync_map == POOL_SYNC_MAP_EMPTY)
 				continue;
 
+			if (use_sync_map == POOL_SYNC_MAP_IS_VALID && !pool_is_set_sync_map(i))
+			{
+				continue;
+			}
+
 			if (pool_read(CONNECTION(backend, i), &kind, sizeof(kind)))
 				return POOL_END;
 
@@ -1627,7 +1710,16 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 					 */
 					if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) == 'T')
 					{
-						pool_set_writing_transaction();
+						/* However, if the query is "SET TRANSACTION READ ONLY" or its variant,
+						 * don't set it.
+						 */
+						if (!pool_is_transaction_read_only(node))
+						{
+							ereport(DEBUG1,
+									(errmsg("not SET TRANSACTION READ ONLY")));
+
+							pool_set_writing_transaction();
+						}
 					}
 
 					/*
@@ -1791,6 +1883,8 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 {
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_STATUS status;
+	char kind = ' ';
+	char *name = "";
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
@@ -1799,254 +1893,44 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 	status = SimpleForwardToFrontend('3', frontend, backend);
 
 	/* Remove the target message */
-	if (session_context->uncompleted_message)
+	if (STREAM)
 	{
-		pool_remove_sent_message(session_context->uncompleted_message->kind,
-								 session_context->uncompleted_message->name);
-		session_context->uncompleted_message = NULL;
+		POOL_PENDING_MESSAGE *pmsg;
+
+		pmsg = pool_pending_message_remove(POOL_CLOSE);
+
+		if (pmsg)
+		{
+			kind = pool_get_close_message_spec(pmsg);
+			name = pool_get_close_message_name(pmsg);
+		}
 	}
 	else
 	{
-		ereport(ERROR,
-				(errmsg("processing CloseComplete, uncompleted message not found")));
+		if (session_context->uncompleted_message)
+		{
+			kind = session_context->uncompleted_message->kind;
+			name = session_context->uncompleted_message->name;
+			session_context->uncompleted_message = NULL;
+		}
+		else
+		{
+			ereport(ERROR,
+					(errmsg("processing CloseComplete, uncompleted message not found")));
+		}
+	}
+	
+	if (kind == ' ')
+	{
+		pool_remove_sent_message(kind, name);
+		ereport(DEBUG1,
+				(errmsg("CloseComplete: remove uncompleted message. kind:%c, name:%s",
+						kind, name)));
 	}
 
 	return status;
 }
 
-POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
-{
-	int i;
-	int len, len1, sendlen;
-	int rows;
-	char *p, *p1, *p2;
-	bool update_or_delete = false;
-	POOL_SESSION_CONTEXT *session_context;
-
-	/* Get session context */
-	session_context = pool_get_session_context(false);
-
-	if (session_context->query_context != NULL)
-	{
-		Node *node = session_context->query_context->parse_tree;
-
-		if (IsA(node, PrepareStmt))
-		{
-			if (session_context->uncompleted_message)
-			{
-				pool_add_sent_message(session_context->uncompleted_message);
-				session_context->uncompleted_message = NULL;
-			}
-		}
-		else if (IsA(node, DeallocateStmt))
-		{
-			char *name;
-			
-			name = ((DeallocateStmt *)node)->name;
-			if (name == NULL)
-			{
-				pool_remove_sent_messages('Q');
-				pool_remove_sent_messages('P');
-			}
-			else
-			{
-				pool_remove_sent_message('Q', name);
-				pool_remove_sent_message('P', name);
-			}
-		}
-		else if (IsA(node, DiscardStmt))
-		{
-			DiscardStmt *stmt = (DiscardStmt *)node;
-
-			if (stmt->target == DISCARD_PLANS)
-			{
-				pool_remove_sent_messages('Q');
-				pool_remove_sent_messages('P');
-			}
-			else if (stmt->target == DISCARD_ALL)
-			{
-				pool_clear_sent_message_list();
-			}
-		}
-		/*
-		 * JDBC driver sends "BEGIN" query internally if
-		 * setAutoCommit(false).  But it does not send Sync message
-		 * after "BEGIN" query.  In extended query protocol,
-		 * PostgreSQL returns ReadyForQuery when a client sends Sync
-		 * message.  Problem is, pgpool can't know the transaction
-		 * state without receiving ReadyForQuery. So we remember that
-		 * we need to send Sync message internally afterward, whenever
-		 * we receive BEGIN in extended protocol.
-		 */
-		else if (IsA(node, TransactionStmt))
-		{
-			TransactionStmt *stmt = (TransactionStmt *) node;
-
-			if (stmt->kind == TRANS_STMT_BEGIN || stmt->kind == TRANS_STMT_START)
-			{
-				int i;
-
-				for (i = 0; i < NUM_BACKENDS; i++)
-				{
-					if (!VALID_BACKEND(i))
-						continue;
-
-					TSTATE(backend, i) = 'T';
-				}
-				pool_unset_writing_transaction();
-				pool_unset_failed_transaction();
-				pool_unset_transaction_isolation();
-			}
-		}
-
-	}
-
-	pool_read(MASTER(backend), &len, sizeof(len));
-
-	len = ntohl(len);
-	len -= 4;
-	len1 = len;
-
-	p = pool_read2(MASTER(backend), len);
-	if (p == NULL)
-		return POOL_END;
-	p1 = palloc(len);
-	memcpy(p1, p, len);
-
-	rows = extract_ntuples(p);
-
-	/*
-	 * Save number of affected tuples of master node.
-	 */
-	session_context->ntuples[MASTER_NODE_ID] = rows;
-
-	if (strstr(p, "UPDATE") || strstr(p, "DELETE"))
-		update_or_delete = true;
-
-	for (i=0;i<NUM_BACKENDS;i++)
-	{
-		if (!IS_MASTER_NODE_ID(i))
-		{
-			if (!VALID_BACKEND(i))
-			{
-				session_context->ntuples[i] = -1;
-				continue;
-			}
-
-			pool_read(CONNECTION(backend, i), &len, sizeof(len));
-
-			len = ntohl(len);
-			len -= 4;
-
-			p = pool_read2(CONNECTION(backend, i), len);
-			if (p == NULL)
-				return POOL_END;
-
-			if (len != len1)
-			{
-				ereport(DEBUG1,
-					(errmsg("processing command complete"),
-						errdetail("length does not match between backends master(%d) %d th backend(%d)",
-							   len, i, len1)));
-			}
-
-			int n = extract_ntuples(p);
-
-			/*
-			 * Save number of affected tuples.
-			 */
-			session_context->ntuples[i] = n;
-
-			/*
-			 * if we are in the parallel mode, we have to sum up the number
-			 * of affected rows
-			 */
-			if (PARALLEL_MODE && is_parallel_table && update_or_delete)
-			{
-				rows += n;
-			}
-			else
-			{
-				if (rows != n)
-				{
-					/*
-					 * Remember that we have different number of UPDATE/DELETE
-					 * affected tuples in backends.
-					 */
-					session_context->mismatch_ntuples = true;
-				}
-			}
-		}
-	}
-
-	if (session_context->mismatch_ntuples)
-	{
-		char msgbuf[128];
-
-		String *msg = init_string("pgpool detected difference of the number of inserted, updated or deleted tuples. Possible last query was: \"");
-		string_append_char(msg, query_string_buffer);
-		string_append_char(msg, "\"");
-		pool_send_error_message(frontend, MAJOR(backend),
-								"XX001", msg->data, "",
-								"check data consistency between master and other db node",  __FILE__, __LINE__);
-		ereport(LOG,
-			(errmsg("%s", msg->data)));
-		free_string(msg);
-
-		msg = init_string("CommandComplete: Number of affected tuples are:");
-
-		for (i=0;i<NUM_BACKENDS;i++)
-		{
-			snprintf(msgbuf, sizeof(msgbuf), " %d", session_context->ntuples[i]);
-			string_append_char(msg, msgbuf);
-		}
-		ereport(LOG,
-			(errmsg("processing command complete"),
-				 errdetail("%s", msg->data)));
-
-		free_string(msg);
-	}
-	else
-	{
-		if (PARALLEL_MODE && is_parallel_table && update_or_delete)
-		{
-			char tmp[32];
-
-			strncpy(tmp, p1, 7);
-			sprintf(tmp+7, "%d", rows);
-
-			p2 = pstrdup(tmp);
-
-			pfree(p1);
-			p1 = p2;
-			len1 = strlen(p2) + 1;
-		}
-
-		pool_write(frontend, "C", 1);
-		sendlen = htonl(len1+4);
-		pool_write(frontend, &sendlen, sizeof(sendlen));
-		pool_write_and_flush(frontend, p1, len1);
-	}
-
-	/* Save the received result to buffer for each kind */
-	if (pool_config->memory_cache_enabled)
-	{
-		if (pool_is_cache_safe() && !pool_is_cache_exceeded())
-		{
-			memqcache_register('C', frontend, p1, len1);
-		}
-	}
-
-	pfree(p1);
-
-	if (pool_is_doing_extended_query_message())
-	{
-		pool_set_query_state(session_context->query_context, POOL_EXECUTE_COMPLETE);
-	}
-
-	return POOL_CONTINUE;
-}
-
 POOL_STATUS ParameterDescription(POOL_CONNECTION *frontend,
 								 POOL_CONNECTION_POOL *backend)
 {
@@ -2373,6 +2257,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		char *query;
 		Node *node;
 		List *parse_tree_list;
+		bool error;
 
 		case 'X':	/* Terminate */
 			if(contents)
@@ -2398,29 +2283,39 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		case 'P':	/* Parse */
 			allow_close_transaction = 0;
 			pool_set_doing_extended_query_message();
+#ifdef NOT_USED
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
+#endif
 			status = Parse(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'B':	/* Bind */
 			pool_set_doing_extended_query_message();
+#ifdef NOT_USED
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
+#endif
 			status = Bind(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'C':	/* Close */
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
 			status = Close(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'D':	/* Describe */
 			pool_set_doing_extended_query_message();
+#ifdef NOT_USED
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
+#endif
 			status = Describe(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'S':  /* Sync */
@@ -2430,6 +2325,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			if (!pool_is_query_in_progress())
 				pool_set_query_in_progress();
 			status = SimpleForwardToBackend(fkind, frontend, backend, len, contents);
+			pool_unset_pending_response();
 			break;
 
 		case 'F':	/* FunctionCall */
@@ -2440,7 +2336,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			query = "INSERT INTO foo VALUES(1)";
 			MemoryContext old_context = MemoryContextSwitchTo(query_context->memory_context);
 
-			parse_tree_list = raw_parser(query);
+			parse_tree_list = raw_parser(query, &error);
 			node = (Node *) lfirst(list_head(parse_tree_list));
 			pool_start_query(query_context, query, strlen(query) + 1, node);
 
@@ -2498,7 +2394,16 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 	{
 		if (pool_is_query_in_progress())
 			pool_unset_query_in_progress();
-		return POOL_CONTINUE;
+
+		/*
+		 * Check if If we have pending data in backend connection cache. If we
+		 * do, it is likely that a sync message has been sent to backend and the
+		 * backend replied back to us. So we need to process it.
+		 */
+		if (is_backend_cache_empty(backend))
+		{
+			return POOL_CONTINUE;
+		}
 	}
 
 	if (pool_is_skip_reading_from_backends())
@@ -2545,19 +2450,22 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 			case '1':	/* ParseComplete */
 				status = ParseComplete(frontend, backend);
 				pool_set_command_success();
-				pool_unset_query_in_progress();
+				if (REPLICATION||RAW_MODE)
+					pool_unset_query_in_progress();
 				break;
 
 			case '2':	/* BindComplete */
 				status = BindComplete(frontend, backend);
 				pool_set_command_success();
-				pool_unset_query_in_progress();
+				if (REPLICATION||RAW_MODE)
+					pool_unset_query_in_progress();
 				break;
 
 			case '3':	/* CloseComplete */
 				status = CloseComplete(frontend, backend);
 				pool_set_command_success();
-				pool_unset_query_in_progress();
+				if (REPLICATION||RAW_MODE)
+					pool_unset_query_in_progress();
 				break;
 
 			case 'E':	/* ErrorResponse */
@@ -2576,7 +2484,7 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 			case 'C':	/* CommandComplete */				
 				status = CommandComplete(frontend, backend);
 				pool_set_command_success();
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
@@ -2586,25 +2494,31 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 			case 'I':	/* EmptyQueryResponse */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
+				/* Empty query response message should be treated same as
+				 * Command complete message. When we receive the Command
+				 * complete message, we unset the query in progress flag if
+				 * operated in streaming replication mode. So we unset the
+				 * flag as well. See bug 190 for more details.
+				 */
 				if (pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
 			case 'T':	/* RowDescription */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
 			case 'n':	/* NoData */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
 			case 's':	/* PortalSuspended */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
@@ -2739,20 +2653,12 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 	char *string = NULL;
 	int len;
 	int i;
-	DistDefInfo *info = NULL;
 
 #ifdef DEBUG
 	int j = 0;
 	char buf[1024];
 #endif
 
-	if (copyin && pool_config->parallel_mode == TRUE)
-	{
-		info = pool_get_dist_def_info(MASTER_CONNECTION(backend)->sp->database,
-									  copy_schema,
-									  copy_table);
-	}
-
 	for (;;)
 	{
 		if (copyin)
@@ -2760,86 +2666,20 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 			if (MAJOR(backend) == PROTO_MAJOR_V3)
 			{
 				char kind;
-				int sendlen;
-				char *p, *p1;
+				char *contents = NULL;
 
 				pool_read(frontend, &kind, 1);
 
-				if (info && kind == 'd')
-				{
-					int id;
-					pool_read(frontend, &sendlen, sizeof(sendlen));
+				ereport(DEBUG1,
+					(errmsg("copy data rows"),
+						 errdetail("read kind from frontend %c(%02x)", kind, kind)));
 
-					len = ntohl(sendlen) - 4;
+				pool_read(frontend, &len, sizeof(len));
+				len = ntohl(len) - 4;
+				if (len > 0)
+					contents = pool_read2(frontend, len);
 
-					if (len <= 0)
-						return POOL_CONTINUE;
-
-					p = pool_read2(frontend, len);
-					if (p == NULL)
-                        ereport(ERROR,
-                            (errmsg("unable to copy data rows"),
-                                 errdetail("read on backend node failed")));
-
-					/* copy end ? */
-					if (len == 3 && memcmp(p, "\\.\n", 3) == 0)
-					{
-						for (i=0;i<NUM_BACKENDS;i++)
-						{
-							if (VALID_BACKEND(i))
-							{
-								pool_write(CONNECTION(backend, i), &kind, 1);
-								pool_write(CONNECTION(backend, i), &sendlen, sizeof(sendlen));
-								pool_write(CONNECTION(backend, i), p, len);
-							}
-						}
-					}
-					else
-					{
-						p1 = parse_copy_data(p, len, copy_delimiter, info->dist_key_col_id);
-
-						if (!p1)
-	                         ereport(ERROR,
-                                (errmsg("unable to copy data rows"),
-                                     errdetail("unable to parse data")));
-   
-						else if (strcmp(p1, copy_null) == 0)
-                            ereport(ERROR,
-                                    (errmsg("unable to copy data rows"),
-                                     errdetail("key parameter is NULL")));
-
-						id = pool_get_id(info, p1);
-
-						ereport(DEBUG1,
-							(errmsg("copy data rows"),
-								 errdetail("copying id: %d", id)));
-						pfree(p1);
-						if (id < 0 || !VALID_BACKEND(id))
-                            ereport(FATAL,
-								(return_code(1),
-                                    errmsg("unable to copy data rows"),
-                                     errdetail("pool_get_id returns invalid id: %d", id)));
-                                    
-
-                        pool_write(CONNECTION(backend, id), &kind, 1);
-                        pool_write(CONNECTION(backend, id), &sendlen, sizeof(sendlen));
-                        pool_write_and_flush(CONNECTION(backend, id), p, len);
-					}
-				}
-				else
-				{
-					char *contents = NULL;
-					ereport(DEBUG1,
-						(errmsg("copy data rows"),
-							 errdetail("read kind from frontend %c(%02x)", kind, kind)));
-
-					pool_read(frontend, &len, sizeof(len));
-					len = ntohl(len) - 4;
-					if (len > 0)
-						contents = pool_read2(frontend, len);
-
-					SimpleForwardToBackend(kind, frontend, backend, len, contents);
-				}
+				SimpleForwardToBackend(kind, frontend, backend, len, contents);
 
 				/* CopyData? */
 				if (kind == 'd')
@@ -3169,7 +3009,7 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 				(errmsg("parse before bind"),
 					 errdetail("waiting for backend %d completing parse", i)));
 
-			pool_extended_send_and_wait(qc, "P", len, contents, 1, i);
+			pool_extended_send_and_wait(qc, "P", len, contents, 1, i, false);
 		}
 		else
 		{
@@ -3323,21 +3163,71 @@ static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *
 }
 
 /*
- * Extract the number of tuples from CommandComplete message
+ * flatten_set_variable_args
+ *		Given a parsenode List as emitted by the grammar for SET,
+ *		convert to the flat string representation used by GUC.
+ *
+ * We need to be told the name of the variable the args are for, because
+ * the flattening rules vary (ugh).
+ *
+ * The result is NULL if args is NIL (ie, SET ... TO DEFAULT), otherwise
+ * a palloc'd string.
  */
-static int extract_ntuples(char *message)
+static char *
+flatten_set_variable_args(const char *name, List *args)
 {
-	char *rows;
+	StringInfoData buf;
+	ListCell   *l;
 
-	if ((rows = strstr(message, "UPDATE")) || (rows = strstr(message, "DELETE")))
-		rows +=7;
-	else if ((rows = strstr(message, "INSERT")))
+	/* Fast path if just DEFAULT */
+	if (args == NIL)
+		return NULL;
+
+	initStringInfo(&buf);
+
+	/*
+	 * Each list member may be a plain A_Const node, or an A_Const within a
+	 * TypeCast; the latter case is supported only for ConstInterval arguments
+	 * (for SET TIME ZONE).
+	 */
+	foreach(l, args)
 	{
-		rows += 7;
-		while (*rows && *rows != ' ') rows++;
-	}
-	else
-		return 0;
+		Node	   *arg = (Node *) lfirst(l);
+		char	   *val;
+		A_Const    *con;
 
-	return atoi(rows);
+		if (l != list_head(args))
+			appendStringInfoString(&buf, ", ");
+	
+		if (IsA(arg, TypeCast))
+		{
+			TypeCast   *tc = (TypeCast *) arg;
+			arg = tc->arg;
+		}
+		
+		if (!IsA(arg, A_Const))
+			elog(ERROR, "unrecognized node type: %d", (int) nodeTag(arg));
+		con = (A_Const *) arg;
+	
+		switch (nodeTag(&con->val))
+		{
+			case T_Integer:
+				appendStringInfo(&buf, "%ld", intVal(&con->val));
+				break;
+			case T_Float:
+				/* represented as a string, so just copy it */
+				appendStringInfoString(&buf, strVal(&con->val));
+				break;
+			case T_String:
+				val = strVal(&con->val);
+				appendStringInfoString(&buf, val);
+				break;
+			default:
+				ereport(ERROR, (errmsg("unrecognized node type: %d",
+					 (int) nodeTag(&con->val))));
+				break;
+		}
+	}
+	
+	return buf.data;
 }
